from collections.abc import Callable, Iterable

from plado.datalog.evaluator.query_tree import (
    EqualityPredicateNode,
    InequalityPredicateNode,
    InnerNode,
    NumericConditionNode,
    ProjectionNode,
    QNode,
    QueryTreeTransformer,
    SelectNotPredicateNode,
    SelectPredicateNode,
    is_contained,
)
from plado.datalog.numeric import NumericConstraint


class PredicateInserter(QueryTreeTransformer):
    def __init__(self, variables: list[int], predicate: Callable[[QNode], QNode]):
        self.variables: list[int] = variables
        self.predicate: Callable[[QNode], QNode] = predicate

    def visit_generic(self, node: QNode) -> QNode:
        assert is_contained(self.variables, node.get_argument_map())
        return self.predicate(node)

    def visit_join(self, node: QNode) -> QNode:
        if is_contained(self.variables, node.left.get_argument_map()):
            return node.__class__(node.left.accept(self), node.right, node.cost)
        if is_contained(self.variables, node.right.get_argument_map()):
            return node.__class__(node.left, node.right.accept(self), node.cost)
        assert is_contained(self.variables, node.get_argument_map())
        return self.predicate(node)

    def visit_difference(self, node: QNode) -> QNode:
        return node.__class__(node.left.accept(self), node.right, node.cost)

    def visit_product(self, node: QNode) -> QNode:
        return self.visit_join(node)


def _make_predicate_node(
    variable_id: int, value_ref: int, predicate_class: type[QNode]
):
    def create(node: QNode) -> QNode:
        return predicate_class(node, variable_id, value_ref)

    return create


def _insert_predicate_node(
    node: QNode, variables: list[int], value_ref: int, predicate_class: type[QNode]
) -> QNode:
    assert len(variables) >= 1 and len(variables) <= 2
    inserter: PredicateInserter = PredicateInserter(
        variables, _make_predicate_node(variables[0], value_ref, predicate_class)
    )
    return node.accept(inserter)


def _insert_equality_predicates(
    root: QNode, equalities: list[tuple[int, int]], predicate_class: type[QNode]
) -> QNode:
    for eq in equalities:
        root = _insert_predicate_node(root, [eq[0], eq[1]], eq[1], predicate_class)
    return root


def _insert_select_predicates(
    root: QNode, selects: list[tuple[int, int]], predicate_class: type[QNode]
) -> QNode:
    for select in selects:
        root = _insert_predicate_node(root, [select[0]], select[1], predicate_class)
    return root


def insert_constraint_predicate(constraint: NumericConstraint, root: QNode) -> QNode:
    def create_node(node: QNode) -> NumericConditionNode:
        return NumericConditionNode(node, constraint)

    return root.accept(
        PredicateInserter(sorted(constraint.expr.get_variables()), create_node)
    )


def insert_projections(node: QNode, args: set[int]) -> QNode:
    if isinstance(node, InnerNode):
        common_args = (
            set(
                node.left.get_argument_map().keys()
                & node.right.get_argument_map().keys()
            )
            | args
        )
        node = node.__class__(
            insert_projections(node.left, common_args),
            insert_projections(node.right, common_args),
            node.cost,
        )
    p_args = args & set(node.get_argument_map().keys())
    if len(p_args) != len(node.get_argument_map()):
        return ProjectionNode(node, p_args)
    return node


def insert_filter_predicates(
    vars_eq: Iterable[tuple[int, int]],
    vars_neq: Iterable[tuple[int, int]],
    obj_eq: Iterable[tuple[int, int]],
    obj_neq: Iterable[tuple[int, int]],
    jt_root: QNode,
) -> QNode:
    """
    Insert all = and != predicates into the join tree.
    """
    jt_root = _insert_equality_predicates(jt_root, vars_eq, EqualityPredicateNode)
    jt_root = _insert_equality_predicates(jt_root, vars_neq, InequalityPredicateNode)
    jt_root = _insert_select_predicates(jt_root, obj_eq, SelectPredicateNode)
    jt_root = _insert_select_predicates(jt_root, obj_neq, SelectNotPredicateNode)
    return jt_root



from plado.datalog.program import Clause
from plado.semantics.task import SimpleCondition


def condition_as_clause(num_relations: int, condition: SimpleCondition, clause: Clause):
    a, b, c = clause.pos_body, clause.neg_body, clause.constraints
    condition.to_datalog(num_relations, clause)
    clause.pos_body = a + clause.pos_body
    clause.neg_body = b + clause.neg_body
    clause.constraints = c + clause.constraints



import itertools
from collections.abc import Iterable

from plado.datalog.evaluator.compiler import (
    FLUENTS,
    RELATIONS,
    compile_interdepending,
    compile_without_dependencies,
)
from plado.datalog.evaluator.filtering import (
    insert_constraint_predicate,
    insert_filter_predicates,
    insert_projections,
)
from plado.datalog.evaluator.join_graph import construct_join_graph
from plado.datalog.evaluator.planner import GreedyOptimizer
from plado.datalog.evaluator.query_tree import QNode
from plado.datalog.numeric import NumericConstraint
from plado.datalog.program import Atom, Clause, Constant, DatalogProgram
from plado.utils import Float, tarjan

Table = set[tuple[int]]
Database = list[Table]
FluentsTable = dict[tuple[int], Float]
FluentsDatabase = list[FluentsTable]


def _cost_function(relations, args, join_relation, join_args):
    return -len(join_args)


class NormalizedClause:
    def __init__(
        self,
        head: Atom,
        num_variables: int,
        positive: Iterable[Atom],
        negative: Iterable[Atom],
        vars_eq: Iterable[tuple[int, int]],
        vars_neq: Iterable[tuple[int, int]],
        obj_eq: Iterable[tuple[int, int]],
        obj_neq: Iterable[tuple[int, int]],
        constraints: Iterable[NumericConstraint],
    ):
        self.num_variables: int = num_variables
        self.head: Atom = head
        self.positive: tuple[Atom] = tuple(positive)
        self.negative: tuple[Atom] = tuple(negative)
        self.vars_eq = tuple(vars_eq)
        self.vars_neq = tuple(vars_neq)
        self.obj_eq = tuple(obj_eq)
        self.obj_neq = tuple(obj_neq)
        self.constrants = tuple(constraints)
        assert all((
            any(varid in atom.get_variables() for atom in self.positive)
            for varid in range(self.num_variables)
        )), "all variables must be positively bounded"

    def __str__(self):
        body = filter(
            lambda x: len(x.strip()) > 0,
            [
                ", ".join((str(a) for a in self.positive)),
                ", ".join(("not " + str(a) for a in self.negative)),
                ", ".join((f"?x{x} = ?x{y}" for x, y in self.vars_eq)),
                ", ".join((f"?x{x} != ?x{y}" for x, y in self.vars_neq)),
                ", ".join((f"?x{x} = {y}" for x, y in self.obj_eq)),
                ", ".join((f"?x{x} != {y}" for x, y in self.obj_neq)),
            ],
        )
        return f"{self.head} =: {', '.join(body)}"


def _generate_query_tree(
    clause: NormalizedClause,
    cost_function=_cost_function,
) -> QNode:
    jg = construct_join_graph(clause.num_variables, clause.positive, clause.negative)
    planner = GreedyOptimizer(cost_function)
    qnode = insert_filter_predicates(
        clause.vars_eq, clause.vars_neq, clause.obj_eq, clause.obj_neq, planner(jg)
    )
    for constraint in clause.constrants:
        qnode = insert_constraint_predicate(constraint, qnode)
    return insert_projections(
        qnode,
        set((arg.id for arg in clause.head.arguments)),
    )


def _get_dependency_graph(
    num_relations: int, clauses: Iterable[NormalizedClause]
) -> list[list[int]]:
    dg = [set() for i in range(num_relations)]
    for clause in clauses:
        head = clause.head.relation_id
        for atom in itertools.chain(clause.positive, clause.negative):
            dg[head].add(atom.relation_id)
    return dg


def _get_dependent_components(dg: list[list[int]]) -> list[list[int]]:
    num_relations = len(dg)
    visited: list[bool] = [False for i in range(num_relations)]
    result: list[list[int]] = []

    def on_scc(scc: list[int]):
        for r in scc:
            visited[r] = True
        result.append(scc)

    def get_successors(r: int) -> list[int]:
        return (rr for rr in dg[r] if not visited[rr])

    for r in range(num_relations):
        if not visited[r]:
            tarjan(r, get_successors, on_scc)

    return result


def _is_stratified(
    num_relations: int, clauses: Iterable[NormalizedClause], components: list[list[int]]
):
    component_idx = [None for i in range(num_relations)]
    for i, c in enumerate(components):
        for r in c:
            component_idx[r] = i
    for clause in clauses:
        for atom in clause.negative:
            if (
                component_idx[atom.relation_id]
                >= component_idx[clause.head.relation_id]
            ):
                return False
    return True


def _get_query_engine_code(
    num_relations: int,
    clauses: list[NormalizedClause],
    cost_function=_cost_function,
):
    dependency_graph = _get_dependency_graph(num_relations, clauses)
    dependent_clauses = _get_dependent_components(dependency_graph)
    assert _is_stratified(num_relations, clauses, dependent_clauses)
    query_trees = list(
        (_generate_query_tree(clause, cost_function) for clause in clauses)
    )
    evaluator_code = [
        f"### num clauses: {len(clauses)}",
        f"### num relations: {num_relations}",
    ] + [
        f"### clause {idx}: {clauses[idx].head} := {tree}"
        for idx, tree in enumerate(query_trees)
    ]
    for group in dependent_clauses:
        clause_idxs = [
            i for i, clause in enumerate(clauses) if clause.head.relation_id in group
        ]
        if len(clause_idxs) == 0:
            continue
        if len(group) > 1 or group[0] in dependency_graph[group[0]]:
            relations = [clauses[idx].head.relation_id for idx in clause_idxs]
            relation_args = [
                tuple((symb.id for symb in clauses[idx].head.arguments))
                for idx in clause_idxs
            ]
            rules = [query_trees[idx] for idx in clause_idxs]
            evaluator_code.extend((
                f"## {clauses[idx].head} := {str(query_trees[idx])}"
                for idx in clause_idxs
            ))
            evaluator_code.append(
                compile_interdepending(relations, relation_args, rules, num_relations)
            )
        else:
            assert len(group) == 1
            for idx in clause_idxs:
                clause = clauses[idx]
                evaluator_code.append(
                    f"## {clauses[idx].head} := {str(query_trees[idx])}"
                )
                evaluator_code.append(
                    compile_without_dependencies(
                        clause.head.relation_id,
                        tuple((arg.id for arg in clause.head.arguments)),
                        query_trees[idx],
                        num_relations,
                    )
                )
    evaluator_code = "\n".join(evaluator_code)
    # print()
    # print()
    # print(evaluator_code)
    # print()
    # print()
    return compile(evaluator_code, "<string>", "exec")


def _extract_eq_atom(
    source: Iterable[Atom],
    atoms: list[Atom],
    variables: list[tuple[int, int]],
    objs: list[tuple[int, int]],
    eq_relation: int,
):
    for atom in source:
        if atom.relation_id == eq_relation:
            assert len(atom.arguments) == 2
            x = atom.arguments[0]
            y = atom.arguments[1]
            if x.is_variable():
                if y.is_variable():
                    variables.append((min(x.id, y.id), max(x.id, y.id)))
                else:
                    objs.append((x.id, y.id))
            elif y.is_variable():
                objs.append((y.id, x.id))
            else:
                assert False
        else:
            atoms.append(atom)


def _normalize_clause(
    clause: Clause, eq_relation: int, object_relation: int
) -> NormalizedClause:
    positive = []
    vars_eq = []
    objs_eq = []
    _extract_eq_atom(clause.pos_body, positive, vars_eq, objs_eq, eq_relation)

    negative = []
    vars_neq = []
    objs_neq = []
    _extract_eq_atom(clause.neg_body, negative, vars_neq, objs_neq, eq_relation)

    num_variables = 0
    for atom in itertools.chain([clause.head], positive, negative):
        variables = atom.get_variables()
        if len(variables) > 0:
            num_variables = max(num_variables, max(variables) + 1)
    for constr in clause.constraints:
        variables = constr.expr.get_variables()
        if len(variables) > 0:
            num_variables = max(num_variables, max(variables) + 1)

    for varid in range(num_variables):
        if not any((varid in atom.get_variables() for atom in positive)):
            positive.append(Atom(object_relation, [Constant(varid, True)]))

    return NormalizedClause(
        clause.head,
        num_variables,
        positive,
        negative,
        vars_eq,
        vars_neq,
        objs_eq,
        objs_neq,
        clause.constraints,
    )


class DatalogEngine:
    def __init__(
        self,
        program: DatalogProgram,
        num_objects: int,
        cost_function=_cost_function,
    ):
        self.num_relations = program.num_relations() + 1
        self.object_relation = program.num_relations()
        self.code = _get_query_engine_code(
            self.num_relations,
            list(
                _normalize_clause(
                    clause, program.equality_relation, self.object_relation
                )
                for clause in program.clauses
            ),
            cost_function,
        )
        self.static_atoms = list(program.trivial_clauses)
        self.static_atoms.extend([
            Atom(self.object_relation, [Constant(obj, False)])
            for obj in range(num_objects)
        ])

    def __call__(
        self, facts: Database, fluents: FluentsDatabase | None = None
    ) -> Database:
        env = {FLUENTS: fluents, RELATIONS: [set(r) for r in facts]}
        env[RELATIONS].append(set())
        for atom in self.static_atoms:
            env[RELATIONS][atom.relation_id].add(
                tuple((arg.id for arg in atom.arguments))
            )
        exec(self.code, env)
        del env[RELATIONS][self.object_relation]
        return env[RELATIONS]



from plado.pddl.boolean_expression import BooleanExpression, Predicate


class DerivedPredicate:
    def __init__(self, predicate: Predicate, condition: BooleanExpression):
        self.predicate: Predicate = predicate
        self.condition: BooleanExpression = condition

    def dump(self, level: int) -> str:
        return (
            "(derived-predicate\n"
            f"{' ' * (level + 1)}  {str(self.predicate)}\n"
            f"{' ' * (level + 1)}  {self.condition.dump(level+2)})"
        )



import itertools
from collections.abc import Callable, Iterable
from typing import Any, TypeVar

T = TypeVar("T")


def tarjan(
    init: T,
    get_successors: Callable[[T], Iterable[T]],
    on_scc: Callable[[list[T]], Any] | None = None,
    stack: list[T] | None = None,
) -> None:
    """
    Run Tarjan's algorithm on the given graph. Notifies on_scc whenever a
    max-SCC is found. stack is the recursive node expansion stack (optional).
    get_successor takes a node (of type T) as input and returns its neighbors
    in the graph. init is the root node from where search is started.
    """

    class NodeInfo:
        def __init__(self, index: int, successors: Iterable[T]):
            self.index = index
            self.backlink = index
            self.successors = iter(successors)

    if stack is None:
        stack = []
    elif len(stack) > 0:
        del stack[0:]

    def noop(stck: list[T]):
        pass

    if on_scc is None:
        on_scc = noop
    stack.append(init)
    infos = {init: NodeInfo(0, get_successors(init))}
    node = init
    backlink = 0
    while True:
        info = infos[node]
        info.backlink = min(info.backlink, backlink)
        try:
            while True:
                succ = next(info.successors)
                succ_info = infos.get(succ, None)
                if succ_info is None:
                    infos[succ] = NodeInfo(len(stack), get_successors(succ))
                    stack.append(succ)
                    node = succ
                    info = infos[succ]
                elif succ_info.index >= 0:
                    info.backlink = min(info.backlink, succ_info.index)
        except StopIteration:
            pass
        index = info.index
        backlink = info.backlink
        if index == backlink:
            on_scc(stack[index:])
            for i in range(index, len(stack)):
                infos[stack[i]].index = -1
            del stack[index:]
            if index == 0:
                break
        node = stack[index - 1]


def reachability_closure(
    init: T, get_successors: Callable[[T], Iterable[T]]
) -> dict[T, list[T]]:
    """
    Computes the transitive reachability closure of all nodes reachable from
    init. Similarly to tarjan, get_successors returns a node's neighbors
    in the graph.
    """
    closure = {}
    stack = []

    def on_scc(scc: list[T]) -> None:
        stack_closure = set()
        for t in scc:
            for s in get_successors(t):
                assert s in scc or (s in closure and s in closure[s])
                stack_closure.update(closure.get(s, set()))
            stack_closure.update(closure.get(t, set([t])))
        for t in scc:
            closure[t] = stack_closure
        for t in stack:
            closure[t] = closure.get(t, set([t])) | stack_closure

    tarjan(init, get_successors, on_scc, stack)
    return {t: list(c) for t, c in closure.items()}



from collections.abc import Iterable

from plado.pddl.arguments import ArgumentDefinition
from plado.pddl.boolean_expression import Atom, BooleanExpression
from plado.pddl.effects import NumericAssignEffect
from plado.pddl.metrics import Metric
from plado.pddl.numeric_expression import NumericConstant

CR = "\n"


class Problem:
    def __init__(
        self,
        name: str,
        domain_name: str,
        objects: Iterable[ArgumentDefinition],
        initial: Iterable[Atom | NumericAssignEffect],
        goal: BooleanExpression,
        goal_reward: NumericConstant | None,
        metric: Metric | None,
    ):
        self.name: str = name
        self.domain_name: str = domain_name
        self.objects: tuple[ArgumentDefinition] = tuple(objects)
        self.initial: tuple[Atom | NumericAssignEffect] = tuple(initial)
        self.goal: BooleanExpression = goal
        self.goal_reward: NumericConstant | None = goal_reward
        self.metric: Metric | None = metric

    def dump(self) -> str:
        return CR.join([
            f"(define (problem {self.name})",
            f"(:domain {self.domain_name})",
            f"(:objects {' '.join((str(o) for o in self.objects))})",
            "(:init",
            CR.join((f"  {x.dump(1)}" for x in self.initial)),
            ")",
            "(:goal",
            f"  {self.goal.dump(1)}",
            ")",
            f"{self.metric.dump(0) if self.metric else ''}",
            ")"
        ])



from collections.abc import Iterable
from typing import Any

from plado.pddl.arguments import Argument, ArgumentDefinition, Substitution
from plado.pddl.numeric_expression import NumericExpression

CR = "\n"


class Predicate:
    def __init__(self, name: str, parameters: Iterable[ArgumentDefinition]):
        self.name: str = name
        self.parameters: tuple[ArgumentDefinition] = tuple(parameters)

    def __call__(self, args: Iterable[Argument]) -> "Atom":
        return Atom(self.name, args)

    def dump(self) -> str:
        return (
            f"({self.name}"
            f"{' ' if len(self.parameters) > 0 else ''}"
            f"{' '.join((str(arg) for arg in self.parameters))})"
        )

    def __str__(self) -> str:
        return self.dump()


class BooleanExpression:
    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        raise NotImplementedError()

    def substitute(self, substitution: Substitution) -> "BooleanExpression":
        raise NotImplementedError()

    def negate(self) -> "BooleanExpression":
        raise NotImplementedError()

    def simplify(self) -> "BooleanExpression":
        raise NotImplementedError()

    def get_free_variables(self) -> set[str]:
        raise NotImplementedError()

    def __and__(self, other: "BooleanExpression") -> "BooleanExpression":
        return Conjunction([self, other])

    def __or__(self, other: "BooleanExpression") -> "BooleanExpression":
        return Disjunction([self, other])

    def __neg__(self) -> "BooleanExpression":
        return Negation(self)

    def dump(self, level: int) -> str:
        raise NotImplementedError()


class Truth(BooleanExpression):
    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_truth(self)

    def substitute(self, substitution: Substitution) -> "Truth":
        return self

    def negate(self) -> BooleanExpression:
        return Falsity()

    def simplify(self) -> "Truth":
        return self

    def get_free_variables(self) -> set[str]:
        return set()

    def dump(self, level: int) -> str:
        return "(and )"


class Falsity(BooleanExpression):
    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_falsity(self)

    def substitute(self, substitution: Substitution) -> "Falsity":
        return self

    def negate(self) -> BooleanExpression:
        return Truth()

    def simplify(self) -> "Falsity":
        return self

    def get_free_variables(self) -> set[str]:
        return set()

    def dump(self, level: int) -> str:
        return "(or )"


class Atom(BooleanExpression):
    def __init__(self, name: str, arguments: Iterable[Argument]):
        self.name: str = name
        self.arguments: tuple[Argument] = tuple(arguments)

    def substitute(self, substitution: Substitution) -> "Atom":
        return Atom(self.name, (arg.apply(substitution) for arg in self.arguments))

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_atom(self)

    def negate(self) -> BooleanExpression:
        return Negation(self)

    def simplify(self) -> "Atom":
        return self

    def get_free_variables(self) -> set[str]:
        return set((arg.name for arg in self.arguments if arg.is_variable()))

    def __str__(self):
        return self.dump(0)

    def dump(self, level: int) -> str:
        return (
            f"({self.name}"
            f"{' ' if len(self.arguments) > 0 else ''}"
            f"{' '.join((str(arg) for arg in self.arguments))})"
        )


class Negation(BooleanExpression):
    def __init__(self, sub_formula: BooleanExpression):
        self.sub_formula: BooleanExpression = sub_formula

    def substitute(self, substitution: Substitution) -> "Negation":
        return Negation(self.sub_formula.substitute(substitution))

    def negate(self) -> BooleanExpression:
        return self.sub_formula

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_negation(self)

    def simplify(self) -> BooleanExpression:
        simplified = self.sub_formula.simplify()
        if isinstance(simplified, Truth):
            return Falsity()
        if isinstance(simplified, Falsity):
            return Truth()
        if simplified is self.sub_formula:
            return self
        return Negation(simplified)

    def get_free_variables(self) -> set[str]:
        return self.sub_formula.get_free_variables()

    def dump(self, level: int) -> str:
        return f"(not {self.sub_formula.dump(level)})"


class BooleanConnector(BooleanExpression):
    OP = None

    def __init__(self, sub_formulas: Iterable[BooleanExpression]):
        self.sub_formulas: tuple[BooleanExpression] = tuple(sub_formulas)

    def substitute(self, substitution: Substitution) -> "BooleanConnector":
        return self.__class__((f.substitute(substitution) for f in self.sub_formulas))

    def inverse_connector(self) -> type[BooleanExpression]:
        raise NotImplementedError()

    def negate(self) -> BooleanExpression:
        return self.inverse_connector()([f.negate() for f in self.sub_formulas])

    def get_free_variables(self) -> set[str]:
        result = set()
        for f in self.sub_formulas:
            result = result | f.get_free_variables()
        return result

    def dump(self, level: int) -> str:
        if len(self.sub_formulas) == 0:
            return "(and)"
        if len(self.sub_formulas) == 1:
            return f"(and {self.sub_formulas[0].dump(level)})"
        return "".join([
            f"({self.OP}{CR}",
            CR.join(
                (" " * (level * 2 + 2) + c.dump(level + 1) for c in self.sub_formulas)
            ),
            "\n",
            " " * (level * 2),
            ")",
        ])


class Conjunction(BooleanConnector):
    OP = "and"

    def inverse_connector(self) -> type[BooleanExpression]:
        return Disjunction

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_conjunction(self)

    def simplify(self) -> BooleanExpression:
        subs = []
        for sub in (sub.simplify() for sub in self.sub_formulas):
            if isinstance(sub, Truth):
                continue
            if isinstance(sub, Falsity):
                return Falsity()
            if isinstance(sub, Conjunction):
                subs.extend(sub.sub_formulas)
            else:
                subs.append(sub)
        if len(subs) == 0:
            return Truth()
        if len(subs) == 1:
            return subs[0]
        return self.__class__(subs)


class Disjunction(BooleanConnector):
    OP = "or"

    def inverse_connector(self) -> type[BooleanExpression]:
        return Conjunction

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_disjunction(self)

    def simplify(self) -> BooleanExpression:
        subs = []
        for sub in (sub.simplify() for sub in self.sub_formulas):
            if isinstance(sub, Truth):
                return Truth()
            if isinstance(sub, Falsity):
                continue
            if isinstance(sub, Disjunction):
                subs.extend(sub.sub_formulas)
            else:
                subs.append(sub)
        if len(subs) == 0:
            return Falsity()
        if len(subs) == 1:
            return subs[0]
        return self.__class__(subs)


class Quantification(BooleanExpression):
    OP = None

    def __init__(
        self, parameters: Iterable[ArgumentDefinition], sub_formula: BooleanExpression
    ):
        self.parameters: tuple[ArgumentDefinition] = tuple(parameters)
        self.sub_formula: BooleanExpression = sub_formula

    def substitute(self, substitution: Substitution) -> "Quantification":
        new_subs: Substitution = {
            x: y
            for (x, y) in substitution.items()
            if all((p.name != x.name) for p in self.parameters)
        }
        return self.__class__(self.parameters, self.sub_formula.substitute(new_subs))

    def inverse_quantifier(self) -> type[BooleanExpression]:
        raise NotImplementedError()

    def negate(self) -> BooleanExpression:
        return self.inverse_quantifier()(self.parameters, self.sub_formula.negate())

    def simplify(self) -> BooleanExpression:
        sub = self.sub_formula.simplify()
        if isinstance(sub, (Falsity, Truth)):
            return sub
        if sub.__class__ == self.__class__:
            sub_parameters = set((p.name for p in sub.parameters))
            new_parameters = [
                p for p in self.parameters if p.name not in sub_parameters
            ]
            new_parameters.extend(sub.parameters)
            referenced_params = sub.sub_formula.get_free_variables()
            new_parameters = [p for p in new_parameters if p.name in referenced_params]
            return self.__class__(new_parameters, sub.sub_formula)
        if self.sub_formula is sub:
            return self
        return self.__class__(self.parameters, sub)

    def get_free_variables(self) -> set[str]:
        params = set((x.name for x in self.parameters))
        return set(
            (x for x in self.sub_formula.get_free_variables() if x not in params)
        )

    def dump(self, level: int) -> str:
        return (
            f"({self.OP} ({' '.join((str(arg) for arg in self.parameters))}){CR}"
            f"{' ' * (level * 2 + 2)}{self.sub_formula.dump(level+1)})"
        )


class Forall(Quantification):
    OP = "forall"

    def inverse_quantifier(self) -> type[BooleanExpression]:
        return Exists

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_forall(self)


class Exists(Quantification):
    OP = "exists"

    def inverse_quantifier(self) -> type[BooleanExpression]:
        return Forall

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_exists(self)


class NumericComparison(BooleanExpression):
    OP = None

    def __init__(self, lhs: NumericExpression, rhs: NumericExpression):
        self.lhs: NumericExpression = lhs
        self.rhs: NumericExpression = rhs

    def inverse_comparator(self) -> type[BooleanExpression]:
        raise NotImplementedError()

    def negate(self) -> BooleanExpression:
        return self.inverse_comparator()(self.lhs, self.rhs)

    def substitute(self, substitution: Substitution) -> "NumericComparison":
        return self.__class__(
            self.lhs.substitute(substitution), self.rhs.substitute(substitution)
        )

    def simplify(self) -> "NumericComparison":
        return self

    def get_free_variables(self) -> set[str]:
        return self.lhs.get_free_variables() | self.rhs.get_free_variables()

    def dump(self, level: int) -> str:
        return f"({self.OP} {self.lhs.dump(level)} {self.rhs.dump(level)})"


class Less(NumericComparison):
    OP = "<"

    def inverse_comparator(self) -> type[BooleanExpression]:
        return GreaterEqual

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_less(self)


class LessEqual(NumericComparison):
    OP = "<="

    def inverse_comparator(self) -> type[BooleanExpression]:
        return Greater

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_less_equal(self)


class Greater(NumericComparison):
    OP = "<"

    def inverse_comparator(self) -> type[BooleanExpression]:
        return LessEqual

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_greater(self)


class GreaterEqual(NumericComparison):
    OP = ">="

    def inverse_comparator(self) -> type[BooleanExpression]:
        return Less

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_greater_equal(self)


class Equals(NumericComparison):
    OP = "="

    def inverse_comparator(self):
        raise ValueError()

    def negate(self) -> BooleanExpression:
        return Negation(self)

    def traverse(self, visitor: "BooleanExpressionVisitor") -> Any:
        return visitor.visit_equals(self)


class BooleanExpressionVisitor:
    def visit_generic(self, formula: BooleanExpression) -> Any:
        raise NotImplementedError()

    def visit_truth(self, formula: Truth) -> Any:
        return self.visit_generic(formula)

    def visit_falsity(self, formula: Falsity) -> Any:
        return self.visit_generic(formula)

    def visit_atom(self, formula: Atom) -> Any:
        return self.visit_generic(formula)

    def visit_negation(self, formula: Negation) -> Any:
        return self.visit_generic(formula)

    def visit_bool_connector(self, formula: BooleanConnector) -> Any:
        return self.visit_generic(formula)

    def visit_conjunction(self, formula: Conjunction) -> Any:
        return self.visit_bool_connector(formula)

    def visit_disjunction(self, formula: Disjunction) -> Any:
        return self.visit_bool_connector(formula)

    def visit_quantification(self, formula: Quantification) -> Any:
        return self.visit_generic(formula)

    def visit_forall(self, formula: Forall) -> Any:
        return self.visit_quantification(formula)

    def visit_exists(self, formula: Exists) -> Any:
        return self.visit_quantification(formula)

    def visit_numeric_condition(self, formula: NumericComparison) -> Any:
        return self.visit_generic(formula)

    def visit_equals(self, formula: Equals) -> Any:
        return self.visit_numeric_condition(formula)

    def visit_less(self, formula: Less) -> Any:
        return self.visit_numeric_condition(formula)

    def visit_less_equal(self, formula: LessEqual) -> Any:
        return self.visit_numeric_condition(formula)

    def visit_greater(self, formula: Greater) -> Any:
        return self.visit_numeric_condition(formula)

    def visit_greater_equal(self, formula: GreaterEqual) -> Any:
        return self.visit_numeric_condition(formula)


class RecursiveBooleanExpressionVisitor(BooleanExpressionVisitor):
    def visit_generic(self, formula: BooleanExpression):
        assert False, "visit_generic should have been unreachable"

    def visit_atomic(self, formula: BooleanExpressionVisitor) -> bool:
        raise NotImplementedError()

    def visit_truth(self, formula: Truth) -> bool:
        return self.visit_atomic(formula)

    def visit_falsity(self, formula: Falsity) -> bool:
        return self.visit_atomic(formula)

    def visit_atom(self, formula: Atom) -> bool:
        return self.visit_atomic(formula)

    def visit_negation(self, formula: Negation) -> bool:
        return formula.sub_formula.traverse(self)

    def visit_bool_connector(self, formula: BooleanConnector) -> bool:
        for e in formula.sub_formulas:
            if e.traverse(self):
                return True
        return False

    def visit_quantification(self, formula: Quantification) -> bool:
        return formula.sub_formula.traverse(self)

    def visit_numeric_condition(self, formula: NumericComparison) -> bool:
        return self.visit_atomic(formula)


class BooleanExpressionTransformer(BooleanExpressionVisitor):
    def visit_generic(self, formula: BooleanExpression) -> BooleanExpression:
        return formula

    def visit_negation(self, formula: Negation) -> BooleanExpression:
        return Negation(formula.sub_formula.traverse(self))

    def visit_bool_connector(self, formula: BooleanConnector) -> BooleanExpression:
        return formula.__class__((
            f for f in (f.traverse(self) for f in formula.sub_formulas) if f is not None
        ))

    def visit_quantification(self, formula: Quantification) -> BooleanExpression:
        return formula.__class__(formula.parameters, formula.sub_formula.traverse(self))



import itertools
from collections.abc import Iterable, Mapping

import plado.datalog.program as datalog
from plado import pddl
from plado.datalog.numeric import (
    Addition,
    Constant,
    Division,
    Fluent,
    Multiplication,
    NumericConstraint,
    NumericExpression,
    Subtraction,
    fluent_iterator,
)
from plado.utils import Float

Grounding = tuple[int]
AtomsStore = list[set[tuple[int]]]
FluentsStore = list[dict[tuple[int], Float]]


def _instantiate_variables(
    args: Iterable[int], variables: Iterable[tuple[int, int]], grounding: Grounding
) -> tuple[int]:
    result = list(args)
    for var, pos in variables:
        result[pos] = grounding[var]
    return tuple(result)


class State:
    def __init__(self, num_predicates: int, num_functions: int):
        self.atoms: list[set[tuple[int]]] = [set() for i in range(num_predicates)]
        self.fluents: list[dict[tuple[int], Float]] = [{} for i in range(num_functions)]

    def duplicate(self) -> "State":
        copy = State(0, 0)
        copy.atoms: list[set[tuple[int]]] = [set(x) for x in self.atoms]
        copy.fluents: list[dict[tuple[int], Float]] = [dict(x) for x in self.fluents]
        return copy


class Atom:
    def __init__(
        self, predicate: int, args: Iterable[int], variables: Iterable[tuple[int, int]]
    ):
        self.predicate: int = predicate
        self.args: tuple[int] = tuple(args)
        self.variables: tuple[int, int] = tuple(variables)

    def _ground_args(self, grounding: Grounding) -> tuple[int]:
        return _instantiate_variables(self.args, self.variables, grounding)

    def instantiate(self, grounding: Grounding) -> "Atom":
        return Atom(self.predicate, self._ground_args(grounding), [])

    def evaluate(self, grounding: Grounding, atoms: AtomsStore) -> bool:
        return self._ground_args(grounding) in atoms[self.predicate]

    def to_datalog(self) -> datalog.Atom:
        arguments = [datalog.Constant(arg, False) for arg in self.args]
        for var, pos in self.variables:
            arguments[pos] = datalog.Constant(var, True)
        return datalog.Atom(self.predicate, arguments)

    def __str__(self):
        args = [str(arg) for arg in self.args]
        for var, pos in self.variables:
            args[pos] = f"?x{var}"
        sep = "" if len(self.args) == 0 else " "
        return f"(P{self.predicate}{sep}{' '.join(args)})"


class SimpleCondition:
    def __init__(
        self,
        atoms: Iterable[Atom],
        negated_atoms: Iterable[Atom],
        constraints: Iterable[NumericConstraint],
    ):
        self.atoms: tuple[Atom] = tuple(atoms)
        self.negated_atoms: tuple[Atom] = tuple(negated_atoms)
        self.constraints: tuple[NumericConstraint] = tuple(constraints)

    def evaluate(self, grounding: Grounding, state: State) -> bool:
        for atom in self.atoms:
            if not atom.evaluate(grounding, state.atoms):
                return False
        for natom in self.negated_atoms:
            if natom.evaluate(grounding, state.atoms):
                return False
        for constraint in self.constraints:
            if not constraint.evaluate(grounding, state.fluents):
                return False
        return True

    def get_variables(self) -> list[int]:
        return sorted(
            set(
                itertools.chain(
                    itertools.chain.from_iterable((
                        (x for x, _ in atom.variables)
                        for atom in itertools.chain(self.atoms, self.negated_atoms)
                    )),
                    itertools.chain.from_iterable(
                        (c.expr.get_variables() for c in self.constraints)
                    ),
                )
            )
        )

    def to_datalog(self, num_predicates: int, clause: datalog.Clause) -> None:
        clause.pos_body = list((atom.to_datalog() for atom in self.atoms))
        clause.neg_body = tuple((atom.to_datalog() for atom in self.negated_atoms))
        clause.constraints = self.constraints
        for c in self.constraints:
            for fluent in fluent_iterator(c.expr):
                clause.pos_body.append(
                    Atom(
                        fluent.function_id + num_predicates,
                        fluent.args,
                        fluent.variables,
                    ).to_datalog()
                )
        clause.pos_body = tuple(clause.pos_body)

    def __str__(self) -> str:
        co = [str(a) for a in self.atoms]
        co.extend([f"not {a}" for a in self.negated_atoms])
        co.extend([str(c) for c in self.constraints])
        return " and ".join(co)


class AtomicEffect:
    def instantiate(self, grounding: Grounding, state: State) -> "AtomicEffect":
        raise NotImplementedError()

    def apply(self, state: State) -> None:
        raise NotImplementedError()


class AddEffect(AtomicEffect):
    def __init__(self, atom: Atom):
        self.atom: Atom = atom

    def instantiate(self, grounding: Grounding, state: State) -> "AddEffect":
        return AddEffect(self.atom.instantiate(grounding))

    def apply(self, state: State) -> None:
        assert len(self.atom.variables) == 0
        state.atoms[self.atom.predicate].add(self.atom.args)


class DelEffect(AtomicEffect):
    def __init__(self, atom: Atom):
        self.atom: Atom = atom

    def instantiate(self, grounding: Grounding, state: State) -> "DelEffect":
        return DelEffect(self.atom.instantiate(grounding))

    def apply(self, state: State) -> None:
        assert len(self.atom.variables) == 0
        try:
            state.atoms[self.atom.predicate].remove(self.atom.args)
        except KeyError:
            pass


class NumericEffect(AtomicEffect):
    def __init__(self, fluent: Fluent, expr: NumericExpression):
        self.fluent: Fluent = fluent
        self.expr: NumericExpression = expr

    def instantiate(self, grounding: Grounding, state: State) -> "NumericEffect":
        value = self.expr.evaluate(grounding, state.fluents)
        args = _instantiate_variables(
            self.fluent.args, self.fluent.variables, grounding
        )
        return NumericEffect(Fluent(self.fluent.function_id, args, []), Constant(value))

    def apply(self, state: State) -> None:
        assert len(self.fluent.variables) == 0
        assert isinstance(self.expr, Constant)
        state.fluents[self.fluent.function_id][self.fluent.args] = self.expr.value


class ConditionalEffect:
    def __init__(
        self,
        idx: int,
        parameters: Iterable[int],
        cond: SimpleCondition,
        eff: AtomicEffect,
    ):
        self.idx: int = idx
        self.parameters: tuple[int] = parameters
        self.condition: SimpleCondition = cond
        self.effect: AtomicEffect = eff


class ProbabilisticEffect:
    def __init__(
        self,
        parameters: Iterable[int],
        outcomes: Iterable[tuple[NumericExpression, tuple[ConditionalEffect]]],
    ):
        self.parameters: tuple[int] = tuple(parameters)
        self.outcomes: tuple[tuple[NumericExpression, tuple[ConditionalEffect]]] = (
            tuple(outcomes)
        )
        self.num_effects: int = sum((len(outs) for _, outs in self.outcomes))


class ActionEffect:
    def __init__(self, effs: Iterable[ProbabilisticEffect], num_atomic_effects: int):
        self.effects: tuple[ProbabilisticEffect] = tuple(effs)
        self.num_atomic_effects: int = num_atomic_effects


class Action:
    def __init__(
        self,
        name: str,
        parameters: int,
        precondition: SimpleCondition,
        effect: ActionEffect,
    ):
        self.name: str = name
        self.parameters: int = parameters
        self.precondition: SimpleCondition = precondition
        self.effect: ActionEffect = effect


class DerivedPredicate:
    def __init__(self, idx: int, condition: SimpleCondition):
        self.id = idx
        self.condition: SimpleCondition = condition


class ArgReplacer:
    def __init__(
        self,
        pred_ids: Mapping[str, int],
        func_ids: Mapping[str, int],
        obj_ids: Mapping[str, int],
        param_ids: Mapping[str, int],
    ):
        self.pred_ids: Mapping[str, int] = pred_ids
        self.func_ids: Mapping[str, int] = func_ids
        self.objs: Mapping[str, int] = obj_ids
        self.params: Mapping[str, int] = param_ids

    def __call__(self, atom) -> Atom | Fluent:
        args = [None for _ in range(len(atom.arguments))]
        variables = []
        for pos, arg in enumerate(atom.arguments):
            if arg.is_variable():
                variables.append((self.params[arg.name], pos))
            else:
                args[pos] = self.objs[arg.name]
        if isinstance(atom, pddl.FunctionCall):
            return Fluent(self.func_ids[atom.name], args, variables)
        return Atom(self.pred_ids[atom.name], args, variables)

    def duplicate(self) -> "ArgReplacer":
        return ArgReplacer(
            self.pred_ids,
            self.func_ids,
            self.objs,
            {**self.params},
        )


def normalize_expression(
    expr: pddl.NumericExpression, args: ArgReplacer
) -> NumericExpression:
    class Normalizer(pddl.NumericExpressionVisitor):
        def visit_generic(self, expr: pddl.NumericExpression) -> None:
            raise ValueError()

        def visit_constant(self, expr: pddl.NumericConstant) -> Constant:
            return Constant(Float(expr.value_str))

        def visit_function_call(self, expr: pddl.FunctionCall) -> Fluent:
            return args(expr)

        def visit_negation(self, expr: pddl.UnaryNegation) -> Subtraction:
            return Subtraction(Constant(0), expr.sub_expression.traverse(self))

        def visit_sum(self, expr: pddl.Sum) -> Addition:
            return Addition(expr.lhs.traverse(self), expr.rhs.traverse(self))

        def visit_subtraction(self, expr: pddl.Subtraction) -> Subtraction:
            return Subtraction(expr.lhs.traverse(self), expr.rhs.traverse(self))

        def visit_product(self, expr: pddl.Product) -> Multiplication:
            return Multiplication(expr.lhs.traverse(self), expr.rhs.traverse(self))

        def visit_division(self, expr: pddl.Division) -> Division:
            return Division(expr.lhs.traverse(self), expr.rhs.traverse(self))

    return expr.traverse(Normalizer())


def normalize_condition(
    cond: pddl.BooleanExpression, args: ArgReplacer
) -> SimpleCondition:
    if isinstance(cond, pddl.Truth):
        return SimpleCondition([], [], [])
    conjuncts = None
    if not isinstance(cond, pddl.Conjunction):
        conjuncts = [cond]
    else:
        conjuncts = cond.sub_formulas
    atoms = []
    neg_atoms = []
    constraints = []
    for x in conjuncts:
        if isinstance(x, pddl.Atom):
            atoms.append(args(x))
        elif isinstance(x, pddl.Negation):
            assert isinstance(x.sub_formula, pddl.Atom)
            neg_atoms.append(args(x.sub_formula))
        else:
            assert isinstance(x, pddl.NumericComparison)
            lhs = normalize_expression(x.lhs, args)
            rhs = normalize_expression(x.rhs, args)
            op = None
            if isinstance(x, pddl.Less):
                op = NumericConstraint.LESS
            elif isinstance(x, pddl.GreaterEqual):
                op = NumericConstraint.GREATER_EQUAL
            elif isinstance(x, pddl.Greater):
                op = NumericConstraint.GREATER
            elif isinstance(x, pddl.LessEqual):
                op = NumericConstraint.LESS_EQUAL
            else:
                op = NumericConstraint.EQUAL
            constraints.append(NumericConstraint(lhs - rhs, op))
    return SimpleCondition(atoms, neg_atoms, constraints)


class Counter:
    def __init__(self, count: int = 0):
        self.count: int = count

    def inc(self) -> int:
        res, self.count = self.count, self.count + 1
        return res


def normalize_atomic_effect(eff: pddl.ActionEffect, args: ArgReplacer) -> AtomicEffect:
    if isinstance(eff, pddl.AtomEffect):
        return AddEffect(args(eff))
    if isinstance(eff, pddl.NegativeEffect):
        return DelEffect(args(eff.atom))
    assert isinstance(eff, pddl.NumericEffect)
    fluent = args(eff.variable)
    expr = normalize_expression(eff.expression, args)
    if isinstance(eff, pddl.NumericAssignEffect):
        return NumericEffect(fluent, expr)
    if isinstance(eff, pddl.ScaleUpEffect):
        return NumericEffect(fluent, fluent * expr)
    if isinstance(eff, pddl.ScaleDownEffect):
        return NumericEffect(fluent, fluent / expr)
    if isinstance(eff, pddl.IncreaseEffect):
        return NumericEffect(fluent, fluent + expr)
    assert isinstance(eff, pddl.DecreaseEffect)
    return NumericEffect(fluent, fluent - expr)


def normalize_conditional_effect(
    eff: pddl.ActionEffect, args: ArgReplacer, counter: Counter
) -> ConditionalEffect:
    parameters = []
    if isinstance(eff, pddl.UniversalEffect):
        args = args.duplicate()
        off = len(args.params)
        args.params.update({p.name: i + off for (i, p) in enumerate(eff.parameters)})
        parameters = [i + off for i in range(len(eff.parameters))]
        eff = eff.effect
    if isinstance(eff, pddl.ConditionalEffect):
        return ConditionalEffect(
            counter.inc(),
            parameters,
            normalize_condition(eff.condition, args),
            normalize_atomic_effect(eff.effect, args),
        )
    return ConditionalEffect(
        counter.inc(),
        parameters,
        SimpleCondition([], [], []),
        normalize_atomic_effect(eff, args),
    )


def normalize_conjunctive_effect(
    eff: pddl.ActionEffect, args: ArgReplacer, counter: Counter
) -> Iterable[ConditionalEffect]:
    if isinstance(eff, pddl.ConjunctiveEffect):
        for ceff in eff.effects:
            yield normalize_conditional_effect(ceff, args, counter)
    else:
        yield normalize_conditional_effect(eff, args, counter)


def normalize_probabilistic_effect(
    eff: pddl.ActionEffect, args: ArgReplacer, counter: Counter
) -> ProbabilisticEffect:
    outcomes = []
    parameters = []
    if isinstance(eff, pddl.UniversalEffect):
        if not isinstance(eff.effect, pddl.ProbabilisticEffect):
            norm = normalize_conditional_effect(eff, args, counter)
            outcomes.append((Constant(1), tuple([norm])))
            return ProbabilisticEffect([], outcomes)
        args = args.duplicate()
        off = len(args.params)
        args.params.update({p.name: i + off for (i, p) in enumerate(eff.parameters)})
        parameters = [i + off for i in range(len(eff.parameters))]
        eff = eff.effect
    if isinstance(eff, pddl.ProbabilisticEffect):
        for out in eff.outcomes:
            outcomes.append((
                normalize_expression(out.probability, args),
                tuple(normalize_conjunctive_effect(out.effect, args, counter)),
            ))
    else:
        assert len(parameters) == 0
        outcomes.append(
            (Constant(1), tuple(normalize_conjunctive_effect(eff, args, counter)))
        )
    return ProbabilisticEffect(parameters, outcomes)


def normalize_effect(eff: pddl.ActionEffect, args: ArgReplacer) -> ActionEffect:
    eff_count = Counter()
    effs = None
    if not isinstance(eff, pddl.ConjunctiveEffect):
        effs = [eff]
    else:
        effs = eff.effects
    return ActionEffect(
        [normalize_probabilistic_effect(eff, args, eff_count) for eff in effs],
        eff_count.count,
    )


def normalize_action(action: pddl.Action, args: ArgReplacer) -> Action:
    args.params: dict[str, int] = {
        arg.name: i for i, arg in enumerate(action.parameters)
    }
    return Action(
        action.name,
        len(args.params),
        normalize_condition(action.precondition, args),
        normalize_effect(action.effect, args),
    )


def normalize_derived_predicate(
    idx: int, predicate: pddl.DerivedPredicate, args: ArgReplacer
) -> DerivedPredicate:
    class ExistsRemover(pddl.BooleanExpressionTransformer):
        def __init__(self, param_ids: dict[str, int]):
            self.param_ids = param_ids

        def visit_exists(self, formula: pddl.Exists) -> pddl.BooleanExpression:
            off = len(self.param_ids)
            self.param_ids.update(
                {arg.name: i + off for i, arg in enumerate(formula.parameters)}
            )
            return formula.sub_formula

    args.params: dict[str, int] = {
        arg.name: i for i, arg in enumerate(predicate.predicate.parameters)
    }
    no_exists = predicate.condition.traverse(ExistsRemover(args.params))
    return DerivedPredicate(
        idx,
        normalize_condition(no_exists, args),
    )


def get_state(
    state: State,
    init: tuple[pddl.Atom | pddl.NumericAssignEffect],
    obj_ids: Mapping[str, int],
    predicate_ids: Mapping[str, int],
    function_ids: Mapping[str, int],
) -> None:
    for x in init:
        if isinstance(x, pddl.Atom):
            assert all((isinstance(a, pddl.ObjectArgument) for a in x.arguments))
            pid = predicate_ids[x.name]
            args = tuple((obj_ids[a.name] for a in x.arguments))
            state.atoms[pid].add(args)
        else:
            fluent = x.variable
            value = x.expression
            fid = function_ids[fluent.name]
            args = tuple((obj_ids[a.name] for a in fluent.arguments))
            assert isinstance(value, pddl.NumericConstant)
            state.fluents[fid][args] = Float(value.value_str)


class FluentPredicateFinder(pddl.RecursiveActionEffectVisitor):
    def __init__(self):
        self.fluent_predicates: set[str] = set()

    def visit_atomic(self, effect: pddl.ActionEffect):
        pass

    def visit_atom_effect(self, effect: pddl.AtomEffect):
        self.fluent_predicates.add(effect.name)


def get_fluent_predicates(domain: pddl.Domain) -> set[str]:
    finder = FluentPredicateFinder()
    for action in domain.actions:
        action.effect.traverse(finder)
    return finder.fluent_predicates


class Task:
    def _partition_predicates(self, domain: pddl.Domain):
        fluent_predicates = get_fluent_predicates(domain)
        predicates = [p for p in domain.predicates if p.name in fluent_predicates]
        dp_predicates: set[str] = set()
        for dp in domain.derived_predicates:
            if dp.predicate.name not in dp_predicates:
                dp_predicates.add(dp.predicate.name)
                predicates.append(dp.predicate)
        assert len(fluent_predicates & dp_predicates) == 0
        predicates.extend((
            p
            for p in domain.predicates
            if p.name not in fluent_predicates and p.name not in dp_predicates
        ))
        self.eq_predicate: int = None
        for i, p in enumerate(predicates):
            if p.name == "=":
                self.eq_predicate: int = i
                break
        if self.eq_predicate is None:
            self.eq_predicate = len(predicates)
            predicates.append(
                pddl.Predicate(
                    "=",
                    [
                        pddl.ArgumentDefinition("?x", "object"),
                        pddl.ArgumentDefinition("?y", "object"),
                    ],
                ),
            )
        self.predicates: tuple[pddl.Predicate] = tuple(predicates)
        self.num_fluent_predicates: int = len(fluent_predicates)
        self.num_derived_predicates: int = len(dp_predicates)
        self.num_static_predicates: int = (
            len(self.predicates)
            - self.num_fluent_predicates
            - self.num_derived_predicates
        )

    def _extract_static_facts(self):
        self.static_facts: tuple[set[tuple[int]]] = tuple(
            self.initial_state.atoms[i]
            for i in range(
                len(self.predicates) - self.num_static_predicates,
                len(self.predicates),
            )
        )
        del self.initial_state.atoms[self.num_fluent_predicates :]

    def __init__(self, domain: pddl.Domain, problem: pddl.Problem):
        self._partition_predicates(domain)
        predicate_ids = {pred.name: i for i, pred in enumerate(self.predicates)}
        self.functions: tuple[pddl.Function] = domain.functions
        function_ids = {func.name: i for i, func in enumerate(self.functions)}
        self.objects: tuple[str] = tuple(
            itertools.chain(
                (c.name for c in domain.constants), (o.name for o in problem.objects)
            )
        )
        object_to_id = {obj: i for i, obj in enumerate(self.objects)}
        self.initial_state: State = State(len(self.predicates), len(self.functions))
        get_state(
            self.initial_state,
            problem.initial,
            object_to_id,
            predicate_ids,
            function_ids,
        )
        self.initial_state.atoms[self.eq_predicate] = set(
            ((o, o) for o in range(len(self.objects)))
        )
        for builtin in ["total-cost", "reward"]:
            if builtin in function_ids:
                self.initial_state.fluents[function_ids[builtin]][tuple([])] = Float(0)
        self._extract_static_facts()
        self.actions: tuple[Action] = tuple((
            normalize_action(
                action, ArgReplacer(predicate_ids, function_ids, object_to_id, None)
            )
            for action in domain.actions
        ))
        self.derived_predicates: tuple[DerivedPredicate] = tuple((
            normalize_derived_predicate(
                predicate_ids[p.predicate.name],
                p,
                ArgReplacer(predicate_ids, function_ids, object_to_id, None),
            )
            for p in domain.derived_predicates
        ))
        self.goal: DerivedPredicate = normalize_derived_predicate(
            None,
            pddl.DerivedPredicate(pddl.Predicate("@goal@", []), problem.goal),
            ArgReplacer(predicate_ids, function_ids, object_to_id, None),
        )

    def create_datalog_program(self) -> datalog.DatalogProgram:
        program = datalog.DatalogProgram()
        for i, pred in enumerate(self.predicates):
            program.add_relation(len(pred.parameters))
        program.equality_relation = self.eq_predicate
        for func in self.functions:
            program.add_relation(len(func.parameters))
        for dp in self.derived_predicates:
            clause = datalog.Clause(
                datalog.Atom(
                    dp.id,
                    [
                        datalog.Constant(i, True)
                        for i in range(len(self.predicates[dp.id].parameters))
                    ],
                ),
                [],
                [],
                [],
            )
            dp.condition.to_datalog(len(self.predicates), clause)
            program.add_clause(clause)
        return program

    def prepare_for_query(
        self, facts: AtomsStore, fluents: FluentsStore
    ) -> tuple[AtomsStore, FluentsStore]:
        assert len(facts) == self.num_fluent_predicates
        assert len(fluents) == len(self.functions)
        efacts = [set(tupls) for tupls in facts]
        efacts.extend((set() for _ in range(self.num_derived_predicates)))
        efacts.extend(self.static_facts)
        efacts.extend((set() for _ in range(len(self.functions))))
        for func, fls in enumerate(fluents):
            efacts[len(self.predicates) + func] = set(fls.keys())
        s = State(0, 0)
        s.atoms = efacts[: len(self.predicates)]
        return efacts, fluents

    def dump_state(self, state: State, end: str = " ") -> str:
        res = []
        for p in range(len(state.atoms)):
            for args in state.atoms[p]:
                res.append(
                    f"({self.predicates[p].name}"
                    f"{' ' if len(self.predicates[p].parameters) > 0 else ''}"
                    f"{' '.join((self.objects[o] for o in args))})"
                )
        for f in range(len(state.fluents)):
            for args in state.fluents[f]:
                res.append(
                    "(= "
                    f"({self.functions[f].name}"
                    f"{' ' if len(self.functions[f].parameters) > 0 else ''}"
                    f"{' '.join((self.objects[o] for o in args))})"
                    f" {state.fluents[f][args]})"
                )
        return end.join(res)

    def dump_action(self, action_id: int, args: tuple[int]) -> str:
        return (
            f"({self.actions[action_id].name}"
            f"{' ' if self.actions[action_id].parameters > 0 else ''}"
            f"{' '.join((self.objects[o] for o in args))})"
        )

    def dump_fact(self, predicate_id: int, args: tuple[int]) -> str:
        return "".join([
            "(",
            self.predicates[predicate_id].name,
            "" if len(args) == 0 else " ",
            " ".join((self.objects[obj] for obj in args)),
            ")",
        ])



class UnionFind:
    """
    Simple union find data structure.
    """

    def __init__(self, n: int = 0):
        self._uf: list[int] = list(range(n))
        self._classes: dict[int, list[int]] = {i: [i] for i in range(n)}

    def __str__(self):
        return "\n".join(
            (f"{i}: {self._const_lookup(i)}" for i in range(len(self._uf)))
        )

    def _const_lookup(self, i: int) -> int:
        assert i < len(self._uf)
        j = i
        k = self._uf[j]
        while j != k:
            j = k
            k = self._uf[j]
        return k

    def _lookup(self, i: int) -> int:
        k = self._const_lookup(i)
        while i != k:
            i, self._uf[i] = self._uf[i], k
        return i

    def resize(self, n: int) -> None:
        if n >= len(self._uf):
            self._classes.update({i: [i] for i in range(len(self._uf), n)})
            self._uf.extend(range(len(self._uf), n))
        else:
            self._uf = self._uf[:n]
            new_classes = {}
            for clas in self._classes.values():
                new_clas = [i for i in clas if i < n]
                if len(new_clas) > 0:
                    min_elem = min(new_clas)
                    new_classes[min_elem] = new_clas
            self._classes = new_classes

    def size(self) -> int:
        return len(self._uf)

    def merge(self, i: int, j: int) -> int:
        i, j = self._lookup(i), self._lookup(j)
        if i == j:
            return i
        i, j = (j, i) if j < i else (i, j)
        self._uf[j] = i
        self._classes[i].extend(self._classes[j])
        del self._classes[j]
        return i

    def are_same(self, i: int, j: int) -> None:
        return self._lookup(i) == self._lookup(j)

    def get_class(self, i: int) -> list[int]:
        return self._classes[self._lookup(i)]

    def __iter__(self):
        return iter(self._classes.values())

    def __getitem__(self, idx: int) -> int:
        return self._lookup(idx)
